#!/usr/bin/perl
# 
# Cobalt Personal Backup
# Will DeHaan, will@cobalt.com/cobalt@guck.net
# (C) Cobalt Networks, Inc. http://cobalt.com/, 2000
#
# There's a reason that 1000+ employee backup-only companies exist.
# The work of one or two hosers can't do justice to disaster recovery.
# Lost email spool?  Missing gifs?  Ok, can do. 
#
#

use strict;
use File::Copy;
use Cobalt::Vsite; # vsite_getgroup

###############
# globals
my $reject_url = '/rejected-feh';

# don't use fileset unless we want to impose an arbitrary file limit
# on users, using fileset with 30000 files is too many arguments
# and the archiving will never even get started
# my $fileset = "* .*"; # we'll copy the current spool into the home dir

# we tar up dot to get around settting an arbitrary limit on the number
# of files we can back up, handles the fringe case where someone decides
# to put 30000 files in their top-level directory thus going way beyond
# the shells limit on argument lists
# this automagically excludes . and ..
# pipe through gzip otherwise uncompressed archive size gets lost
my $tar_cmd = "/bin/tar -cpPf - . --exclude '.restore*' | /bin/gzip -c ";
# my $tar_cmd = "/bin/tar -czpPf - --exclude '.' --exclude '..'";
# my $tar_cmd = "/bin/tar -czpPf - --no-recursion $fileset --exclude '.' --exclude '..'";

###############
# who ami I?
my $user = $ENV{'REMOTE_USER'};
my ($user_id, $user_dir) = (getpwnam($user))[2,7] if (($user =~ /^[a-z0-9][a-z0-9\.\-\_]+$/) && (length($user) < 32));
unless(-d $user_dir) { # such a physical validity test.  What are today's getpwnam exploits?
	die "Authenticated user $user does not have a home directory.  Restore aborting.\n";
}
my $group = Cobalt::Vsite::vsite_getgroup($user);
my $gid = (getgrnam($group))[2];

###############
# render filename, note PRQ for Personal RaQ, .raq is deprecated post RaQ 4's
my($filename, $date_stamp);
chomp($date_stamp = `date +%m-%d-%Y`);
$filename = $user.'-'.$date_stamp.'.tgz'; # Concat so perl doesn't confuse $'s

###############
# save the spools
copy("/home/spool/mail/$user", $user_dir.'/email_spool.'.$date_stamp) || 
        warn "Could not copy mail spool /home/spool/mail/$user: $!";
chown($user_id, $gid, $user_dir.'/email_spool.'.$date_stamp); 

###############
# become the spoon...
chdir($user_dir);

###############
# octet-stream is mishandled by some IE versions
print "Content-type: application/CobaltPersonalBackup\n";
print "Content-disposition: filename=$filename\n\n";
#
# Slam!
open(TAR, "$tar_cmd |") || warn "could not open $tar_cmd: $!";
while(<TAR>) { print; }
close(TAR);

###############
# clean up
unlink($user_dir.'/email_spool.'.$date_stamp);

###############
# No really, that's it folks
exit 0;


