#!/usr/bin/perl
#
# Cobalt Personal Backup, Edition IV
# Will DeHaan, will@cobalt.com/cobalt@guck.net
# (C) 2000 Cobalt Networks, Inc., http://cobalt.com/, 2000
#

use CGI;
use File::Copy;
use strict;
### START HARDCODED PATHS
my $Adminuser = 'admin';
my $Alteradmin = 'alteradmin';
my $MaxFileList = '2000';

my $Debug = (-e "/etc/DEBUG") ? 1 : 0;
$^W = $Debug;
### END HARDCODED PATHS
use Cobalt::Vsite;
use Cobalt::Util qw( update_status );
use Fcntl qw( O_WRONLY O_CREAT O_EXCL );
use File::Copy;
use IPC::Open3;

###############
# who ami I? set euid/egid..
my $q = new CGI;
my $user = $ENV{'REMOTE_USER'};
my ($user_id, $user_dir) = (getpwnam($user))[2,7] if (($user =~ /^[a-z0-9][a-z0-9\.\-\_]+$/) && (length($user) < 32));
unless(-d $user_dir) { # such a physical validity test.  What are today's getpwnam exploits?
	die "Authenticated user $user does not have a home directory.  Restore aborting.\n";
}

my $group = Cobalt::Vsite::vsite_getgroup($user);

my $run_as_root = 0;

if (($user eq $Adminuser) || ($user eq $Alteradmin)) {
	$group ||= $q->param('site');
	$run_as_root = 1;
}

my $type = ($q->param('site_restore') ? 'site' : 'personal');
my $run_flag = ".restore.$type.running";

my $gid = (getgrnam($group))[2];
# so we can move the uploaded file
umask(002);

###############
# parse CGI
my $upload = (defined $q->param('upload')) ? $q->param('upload') : 1;
my $selective = $q->param('selective');
my $selected = $q->param('selected');

# create the run flag and put something in it
open FOO, ">$user_dir/$run_flag" or die "$!\n";

# autoflush
my $old_fh = select FOO;
$| = 1;
select $old_fh;

update_status(\*FOO, [$$, 'START', 0, 1]);

# make sure we have the archive name to return in case this is a selective
# restore
my $archive;
if ($q->param('homedirarchive')) {
	$upload = 0;
	$archive = $user_dir . '/' . $q->param('homedirarchive');
	
	# chown the file, so it won't count against user's quota
	my @owner = (getpwnam('root'))[2, 3];
	chown @owner, $archive;

} elsif ($q->param('archive')) {
	$archive = $q->param('archive');
} elsif ($q->param('backupfile')) {
	$archive = $q->upload('backupfile');
	my $tmpfile = $q->{'.tmpfiles'}->{fileno($archive)}->{name};
	
	if ($archive =~ /([^\/\\]+\.tgz)$/) {
		$archive = $1; # Strip path
	}

	if ($archive) {
		move($$tmpfile, $user_dir . '/' . $archive);
		chmod 0644, ($user_dir . '/' . $archive);	
		$archive = $user_dir ."/" . $archive;
	} else {
		# file upload must have failed
		&error(\*FOO, '[str.backup.upload.failed]');
	}
} else {
	&error(\*FOO, '[str.backup.no.archive.specified]');
}

# now change the uid/gid
$( = $) = $gid;
unless ($run_as_root) { $< = $> = $user_id; }

# now fork off to the background and return to the ui
my @status = ($$, 'START', 0, 1);

$SIG{CHLD} = "IGNORE";

my $pid = fork();

if ($pid) {
	$status[0] = $pid;
	update_status(\*FOO, \@status);
	close FOO;
	
	print STDOUT "Location:  /cgi-bin/.cobalt/turbo_ui/dispatch/restoremenu?" .
		'site_restore=' . $q->param('site_restore') 
		. '&site=' . $group . '&directed=' . $q->param('directed') 
		. "&archive=$archive&upload=$upload\n\n";

	exit 0;
} elsif ($pid != 0) {
	&error(\*FOO, '[str.backup.system.failure]');
}

# clean up for apache, so the parent goes away nicely
# redirect stdin and stdout
open STDIN, "</dev/null";
open STDOUT, ">/dev/null";
open STDERR, ">/dev/null";

# reconnect to the database
my $dbc = Cobalt::Meta::getdb;
$dbc->reset;

# update status again so the correct pid is there
$status[0] = $$;
update_status(\*FOO, \@status);

###############
# become the spoon...
my $base_dir = $user_dir;

if (Cobalt::User::user_authok($user, 'site-adm', $group)) {
		$base_dir = '/home/sites/' . $q->param('site');
}

# admin is doing a restore
if ($user eq $Adminuser || $user eq $Alteradmin) {
	$base_dir = '/';
}

chdir($base_dir);

###############
# Directed Restore
my $directed = $q->param('directed');

# Do not tolerate .../ bs
if ($directed =~ /^\.+\//) { 
	update_status(\*FOO, [$$, '[str.backup.invalid.directed]']);
	exit 0;
 } 
if ($directed =~ /\/\.+\//) { 
	update_status(\*FOO, [$$, '[str.backup.invalid.directed]']);
	exit 0;
} 

if(!$selective || $selected) {
	$directed =~ s/^\///g; # strip leading /'s
	$directed =~ s/\/$//g; # strip trailing /'s
	unless (-d "$base_dir/$directed") {
		system("/bin/mkdir -p \"$base_dir/$directed\" > /dev/null 2>&1");
	}
	if (-d "$base_dir/$directed") { 
		chdir ("$base_dir/$directed");
	} else {
		update_status(\*FOO, [$$, '[str.backup.invalid.directed]']); 
		exit 0;
	}
}

# we move the file above so no need to do this, but I will leave it in
# in case it is useful in the future
# write out the archive if uploaded
# if ($q->param('backupfile') && 0) {
#	my $file = $q->param('backupfile');

#	my $size = (stat($file))[7];

#	@status = ($$, 'UPLOAD', 0, $size);
#	update_status(\*FOO, \@status);

	# store file to $archive in /home/tmp avoiding symlink hack
#	unlink($archive);
#	sysopen(TGZ, $archive, O_WRONLY|O_CREAT|O_EXCL, 0600) || die "Could not open $archive for writing: $!";

#	while(<$file>) { 
#		print TGZ;
#		$status[2] += length($_);
#		update_status(\*FOO, \@status);
#	}
#	close(TGZ);

#	unless (-s $archive) {
#		unlink($archive);
#		update_status(\*FOO, [$$, '[str.backup.insufficient.quota]']);
#		exit 0;
#	}
#}

my $tar_cmd = "LANGUAGE=en_US LC_ALL=en_US /bin/tar -xzpvvRf $archive --no-regex";
my $tar_query = "/bin/tar -tzvf $archive 2>&1";
	
###############
# display selective restore else actually restore...
my (@file, @fileset, %file_sizes, %file_dates, %file_errors, %dir_names);
my $max_test = 0;
if($selective) {
	update_status(\*FOO, [$$, 'INDEX', 0, 1]);

	# parse files from backup archive
	if( not open(ARCH, "$tar_query |")) {
		update_status(\*FOO, [$$, '[str.backup.could.not.open.archive]']);
		exit 0;
	}

	my %dir_map = ();
	my %file_map = ();
	
	open INDEX, ">$user_dir/$run_flag.index" or die "$!\n";

	while(<ARCH>) {
		next if (/(gzip: stdin: decompression OK, trailing garbage ignored|tar: Child returned status 2|tar: Error exit delayed from previous errors)/);

		chop;

		# -rw-r--r-- admin/root        0 2000-09-01 13:55:14 email_spool.20000901
		if (/^[a-zA-Z\-]{10}\s/) {
			@file = split(/\s+/, $_, 6);
			
			&map_name($file[5], \%dir_map, \%file_map);

			print INDEX "$file[5]\t$file[2]\t$file[3] $file[4]\n";

		} elsif ($file[5]) {
			$file_errors{$file[5]} .= $_;
		}
		$max_test++;
		if($max_test > $MaxFileList) {
			update_status(\*FOO, [$$, '[str.backup.too.many.files.for.selective]']); 
			exit 0;
		}
	}

	# use Data::Dumper;
	# open DEBUG, ">/tmp/maps";
	# print DEBUG Dumper \%dir_map;
	# print DEBUG Dumper \%file_map;
	# close DEBUG;

	&write_file_map(\%dir_map, \%file_map, "$user_dir/$run_flag.map");
	close INDEX;
	close(ARCH);

	update_status(\*FOO, [$$, 'SELECT', $archive, $directed]);
} else {
	# Build file index
	my $file_list;
	my $file_index = "$user_dir/.restore.$type.filelist";

	if($selected) {
		my %map = ();
		read_file_map(\%map, "$user_dir/.restore.$type.running.map");

		my $file_count = 0;
		
		$file_list = "--files-from $file_index";
		# param fetch & split, out to file, append syntax w/--from-file...
		my $total_files = $q->param('FILECOUNT'); # this could get long, folks
		my $lame_index = 0;
		if (not open(TMPFILES, "> $file_index")) {
			update_status(\*FOO, [$$, '[str.backup.could.not.write.homedir]']);
			exit 0;
		}

		while($lame_index <= $total_files) {
			my $lame_file = $q->param("FILE$lame_index");
			$lame_index++;
			# don't know why this was needed tar doesn't barf
			# if you give it directories for selective restore
			# --pbaltz 1/29/2001
			# if ($lame_file =~ /\/$/) {
			# $dir_names{"$lame_file"} = $lame_file;
			# next;
			# }
			if ($lame_file) {
				print TMPFILES "$map{$lame_file}\n";
				$file_count++;
			}
		}
		close(TMPFILES);
	
		###############
		# bail if no files were selected
		if (not $file_count) {
			update_status(\*FOO, [$$, '[str.backup.no.files.selected]']);
			exit 0;
		}
	}


	# restore away
	@status = ($$, 'COUNT', 0, 1);

	# update status
	update_status(\*FOO, \@status);

	# `/bin/tar tzf $archive | /usr/bin/wc -l` =~ /(\d+)/;
	# determine the size of the uncompressed tar archive to restore
	# or number of files for a selective restore
	if ($selected) {
		`/bin/cat $file_index  | /usr/bin/wc -l` =~ /(\d+)/;
		@status = ($$, 'RESTORE', 0, $1);
	} else {
		`/bin/gzip -l --quiet $archive` =~ /^\s*(\d+)\s+(\d+)/;
		@status = ($$, 'RESTORE', 0, ($2 ? $2 : $1));
	}

	update_status(\*FOO, \@status);
	local (*IN, *RESTR);
	if (not open3(\*IN, \*RESTR, '', "$tar_cmd $file_list")) {
		update_status(\*FOO, [$$, '[str.backup.could.not.open.archive]']);
		exit 0;
	}
	# open file for file names and info
	open LIST, ">$user_dir/${run_flag}.info" or die "$!\n";
	
	# set bytes to -1 to make the count realistic for selective
	# restore
	my $bytes = -1;

	while(<RESTR>) {
		next if (/(gzip: stdin: decompression OK, trailing garbage ignored|tar: Child returned status 2|tar: Error exit delayed from previous errors)/);
		next unless ($_);
		chop;
		
		# watch for disk full errors
		if (/quota exceeded/) {
			update_status(\*FOO, [$$, '[str.backup.insufficient.quota]']);
			unlink $archive if ($upload);
			exit 0;
		}

		# s/\\\\/\\/g; # tar stores \ as \\, and handles it different on extract
		if (/^block (\d+): [a-zA-Z\-]{10}\s/) {
			@file = split(/\s+/, $_, 8);
			print LIST $file[7] . "\t" . $file[4] . "\n";
		
			$bytes = ($selected ? $bytes + 1 : $1 * 512);	
		
		} elsif (/^block (\d+):/) {
			# the final aka nuls entry
			$bytes = ($selected ? $bytes + 1 : $1 * 512 + 1536);
		} elsif ($file[5]) {
			# /bin/tar: stuff2/chiliasp/admin/web/css/noraq4.css: Not found in archive
			if(/^\/bin\/tar:\s+([\w\d\/\.]+)\:\s+(.+)$/) {
				$file_errors{$1} .= $2;
			}
		}
	
		$status[2] = $bytes;
		update_status(\*FOO, \@status);
	}
	close LIST;
	close RESTR;

	# debatable feature: delete archive after restore if it was uploaded?
	# yes we will do this because they cannot and it could eat up a lot
	# of space
	unlink("$archive");
	
	unlink("$file_index"); # blind deletion. 

	###############
	# backup current; restore mail spool
	my($date_stamp);
	chomp($date_stamp = `date +%Y%m%d`);
	copy("/home/spool/mail/$user", $user_dir.'/email_spool.'.$date_stamp) ||
	warn "Could not copy mail spool /home/spool/mail/$user: $!";
	copy("/home/spool/mail/email_spool", "/home/spool/mail/$user"); # 
	chown($user_id, $gid, $user_dir.'/saved_email.'.$date_stamp);
	# chown($user_id, $gid, "/home/spool/mail/$user");

	# mark that we are done for the review process
	$status[1] = 'DONE';
	$status[2] = $archive;
	$status[3] = $directed;
	update_status(\*FOO, \@status);
}

###############
# that's it, folks
exit 0;

########################## Subs

sub error {
	my ($fh, $user_error) = @_;
	
	$user_error =~ s/\s/\+/g;
	
	update_status($fh, [$$, $user_error]);
	
	print "Location: /cgi-bin/.cobalt/turbo_ui/dispatch/restoremenu?USERERROR=$user_error&site_restore=" . $q->param('site_restore') . "\n\n";
	exit 1;
}

# map directories and files, so we don't need to depend on the browser not
# messing with the value we use to extract from the archive
sub map_name
{
	# open FOO, ">>/tmp/debug";

	my ($name, $dir_map, $file_map) = @_;
	
	my $basename = "";

	# print FOO "original: $name\n";

	# hack off file name if this is not a directory
	if ($name =~ /^(.*\/)([^\/]+)$/) {
		$basename = $2;
		# print FOO "${basename}\n";
		$name = $1;
	}

	my @parts = split /\//, $name;


	my $build = "";
	my $dir_count = scalar(keys %$dir_map);
	my $file_count = scalar(keys %$file_map);
	# print FOO "$name\n$dir_count\t$file_count\n";
	
	# print FOO Dumper \@parts;
	# print FOO "\n";
	for my $part (@parts) {
		my $parent = $build;
		$build .= "$part/";

		# encode this directory path plus store real path, just in
		# case since certain characters aren't allow in hash keys
		if (not exists($$dir_map{$build})) {
			$$dir_map{$build} = $$dir_map{$parent} . ++$dir_count;
		}
	}

	# modify the file map if this is a file and not a directory
	if ($basename ne "") {
		# print FOO "basename: ", $basename, "\n";
		$$file_map{"$name$basename"} = $$dir_map{$build} . '/' . ++$file_count;
	}
	# close FOO;
}

sub write_file_map
{
	my ($dir_map, $file_map, $map_filename) = @_;
	
	open MAP, ">$map_filename" or die "Can't open $map_filename: $!\n";

	for my $key (keys %$dir_map) {
		print MAP "$$dir_map{$key}\t$key\n";
	}

	for my $key (keys %$file_map) {
		print MAP "$$file_map{$key}\t$key\n";
	}

	close MAP;
}

sub read_file_map
{
	my ($map, $map_filename) = @_;

	open MAP, "$map_filename" or die "$!\n";

	for my $line (<MAP>) {
		chomp $line;

		my @parts = split /\t/, $line;

		$$map{$parts[0]} = $parts[1];
	}

	close MAP;
}
